@@ -6,14 +6,37 @@
 Tests for implementations of L{IReactorUDP} and L{IReactorMulticast}.
 """
 
+from __future__ import annotations
 
 import os
+from socket import (
+    AF_INET,
+    AF_INET6,
+    IP_ADD_MEMBERSHIP,
+    IPPROTO_IP,
+    IPPROTO_IPV6,
+    IPV6_JOIN_GROUP,
+    SOCK_DGRAM,
+    AddressFamily,
+    if_nameindex,
+    inet_pton,
+    socket,
+)
 from unittest import skipIf
 
 from twisted.internet import defer, error, interfaces, protocol, reactor, udp
+from twisted.internet.address import IPv4Address, IPv6Address
 from twisted.internet.defer import Deferred, gatherResults, maybeDeferred
+from twisted.internet.error import MulticastJoinError
+from twisted.internet.interfaces import (
+    IMulticastTransport,
+    IReactorMulticast,
+    IReactorTime,
+)
+from twisted.internet.task import deferLater
 from twisted.python import runtime
-from twisted.trial.unittest import TestCase
+from twisted.test.testutils import skipWithoutIPv6
+from twisted.trial.unittest import SkipTest, TestCase
 
 
 class Mixin:
@@ -36,8 +59,9 @@ class Mixin:
 
 
 class Server(Mixin, protocol.DatagramProtocol):
-    packetReceived = None
+    packetReceived: Deferred[None] | None = None
     refused = 0
+    transport: IMulticastTransport
 
     def datagramReceived(self, data, addr):
         self.packets.append((data, addr))
@@ -563,27 +587,66 @@ class ReactorShutdownInteractionTests(TestCase):
         return finished
 
 
+def checkMulticastAvailability(
+    af: AddressFamily, ipproto: int, join: int, group: str, bindto: str
+) -> bool:
+    """
+    Bind a socket to the given network interface and attempt to join a
+    multicast group with the given address family.  This is used to determine
+    whether the local networking stack configuration allows for multicast
+    within the given address family.
+    """
+    s = socket(af, SOCK_DGRAM)
+    s.bind((bindto, 0))
+    packgroup = inet_pton(af, group)
+    packif = inet_pton(af, bindto)
+    try:
+        s.setsockopt(ipproto, join, packgroup + packif)
+    except OSError:
+        return False
+    else:
+        return True
+    finally:
+        s.close()
+
+
 @skipIf(
     not interfaces.IReactorMulticast(reactor, None),
     "This reactor does not support multicast",
 )
+@skipIf(
+    not checkMulticastAvailability(
+        AF_INET, IPPROTO_IP, IP_ADD_MEMBERSHIP, "225.0.0.250", "0.0.0.0"
+    ),
+    "The local networking stack does not enable IPv4 multicast.",
+)
 class MulticastTests(TestCase):
-    if (
-        os.environ.get("INFRASTRUCTURE") == "AZUREPIPELINES"
-        and runtime.platform.isMacOSX()
-    ):
-        skip = "Does not work on Azure Pipelines"
+    """
+    Tests for sending and receiving multicast datagrams.  (This suite supports
+    IPv4, see below for IPv6.)
+    """
 
-    if not interfaces.IReactorMulticast(reactor, None):
-        skip = "This reactor does not support multicast"
+    # Define test addresses as class attributes so they may be parameterized
+    # for IPv6 below.
+    interface: str = "0.0.0.0"
+    expectedInterface: str | int = "0.0.0.0"
+    clientAddress: str = "127.0.0.1"
+    multicastGroup: str = "225.0.0.250"
+    alternateInterface: str | int = "127.0.0.1"
+    interfaceSynonym: str | int = "localhost"
+    invalidGroup: str = "127.0.0.1"
+    wrongAddressFamily: str = "::1"
 
     def setUp(self):
         self.server = Server()
         self.client = Client()
         # multicast won't work if we listen over loopback, apparently
-        self.port1 = reactor.listenMulticast(0, self.server)
-        self.port2 = reactor.listenMulticast(0, self.client)
-        self.client.transport.connect("127.0.0.1", self.server.transport.getHost().port)
+        self.port1 = reactor.listenMulticast(0, self.server, self.interface)
+        self.port2 = reactor.listenMulticast(0, self.client, self.interface)
+        self.client.transport.connect(
+            self.clientAddress,
+            self.server.transport.getHost().port,
+        )
 
     def tearDown(self):
         return gatherResults(
@@ -593,88 +656,84 @@ class MulticastTests(TestCase):
             ]
         )
 
-    def testTTL(self):
-        for o in self.client, self.server:
+    def testTTL(self) -> None:
+        if self.server.transport.addressFamily != AF_INET:
+            raise SkipTest("only IPv4 multicast has TTLs, IPv6 has hops")
+
+        def checkttl(o: Server | Client) -> None:
+            assert IMulticastTransport.providedBy(o.transport)
             self.assertEqual(o.transport.getTTL(), 1)
             o.transport.setTTL(2)
             self.assertEqual(o.transport.getTTL(), 2)
 
-    def _test_loopback(self):
+        checkttl(self.client)
+        checkttl(self.server)
+
+    async def _test_loopback(self) -> None:
         """
         Test that after loopback mode has been set, multicast packets are
         delivered to their sender.
         """
+        group = self.multicastGroup
         self.assertEqual(self.server.transport.getLoopbackMode(), 1)
         addr = self.server.transport.getHost()
-        joined = self.server.transport.joinGroup("225.0.0.250")
 
-        def cbJoined(ignored):
-            d = self.server.packetReceived = Deferred()
-            self.server.transport.write(b"hello", ("225.0.0.250", addr.port))
-            return d
+        await self.server.transport.joinGroup(group)
+        self.server.packetReceived = Deferred()
+        self.server.transport.write(b"hello", (group, addr.port))
+        await self.server.packetReceived
 
-        joined.addCallback(cbJoined)
+        self.assertEqual(len(self.server.packets), 1)
+        self.server.transport.setLoopbackMode(0)
+        self.assertEqual(self.server.transport.getLoopbackMode(), 0)
+        self.server.transport.write(b"hello", (group, addr.port))
 
-        def cbPacket(ignored):
-            self.assertEqual(len(self.server.packets), 1)
-            self.server.transport.setLoopbackMode(0)
-            self.assertEqual(self.server.transport.getLoopbackMode(), 0)
-            self.server.transport.write(b"hello", ("225.0.0.250", addr.port))
+        # This is a recipe for a flaky test, but we need to let the reactor
+        # spin a bit to let the packet through.
+        await deferLater(IReactorTime(reactor), 0)
+        self.assertEqual(len(self.server.packets), 1)
 
-            # This is fairly lame.
-            d = Deferred()
-            reactor.callLater(0, d.callback, None)
-            return d
+    async def test_wrongFamilyInterfaceJoin(self) -> None:
+        """
+        A sensible error will be reported when using the wrong address family
+        literal for the 'interface' argument to 'joinGroup'.
+        """
+        with self.assertRaises(MulticastJoinError):
+            await self.server.transport.joinGroup(
+                self.multicastGroup, interface=self.wrongAddressFamily
+            )
 
-        joined.addCallback(cbPacket)
-
-        def cbNoPacket(ignored):
-            self.assertEqual(len(self.server.packets), 1)
-
-        joined.addCallback(cbNoPacket)
-
-        return joined
-
-    def test_interface(self):
+    async def test_interface(self) -> None:
         """
         Test C{getOutgoingInterface} and C{setOutgoingInterface}.
         """
-        self.assertEqual(self.client.transport.getOutgoingInterface(), "0.0.0.0")
-        self.assertEqual(self.server.transport.getOutgoingInterface(), "0.0.0.0")
+        self.assertEqual(
+            self.client.transport.getOutgoingInterface(), self.expectedInterface
+        )
+        self.assertEqual(
+            self.server.transport.getOutgoingInterface(), self.expectedInterface
+        )
 
-        d1 = self.client.transport.setOutgoingInterface("127.0.0.1")
-        d2 = self.server.transport.setOutgoingInterface("127.0.0.1")
-        result = gatherResults([d1, d2])
+        await self.server.transport.setOutgoingInterface(self.alternateInterface)
+        await self.client.transport.setOutgoingInterface(self.interfaceSynonym)
 
-        def cbInterfaces(ignored):
-            self.assertEqual(self.client.transport.getOutgoingInterface(), "127.0.0.1")
-            self.assertEqual(self.server.transport.getOutgoingInterface(), "127.0.0.1")
+        self.assertEqual(
+            self.client.transport.getOutgoingInterface(),
+            self.alternateInterface,
+        )
+        self.assertEqual(
+            self.server.transport.getOutgoingInterface(),
+            self.alternateInterface,
+        )
 
-        result.addCallback(cbInterfaces)
-        return result
-
-    def _test_joinLeave(self):
+    async def _test_joinLeave(self) -> None:
         """
         Test that multicast a group can be joined and left.
         """
-        d = self.client.transport.joinGroup("225.0.0.250")
-
-        def clientJoined(ignored):
-            return self.client.transport.leaveGroup("225.0.0.250")
-
-        d.addCallback(clientJoined)
-
-        def clientLeft(ignored):
-            return self.server.transport.joinGroup("225.0.0.250")
-
-        d.addCallback(clientLeft)
-
-        def serverJoined(ignored):
-            return self.server.transport.leaveGroup("225.0.0.250")
-
-        d.addCallback(serverJoined)
-
-        return d
+        await self.client.transport.joinGroup(self.multicastGroup)
+        await self.client.transport.leaveGroup(self.multicastGroup)
+        await self.server.transport.joinGroup(self.multicastGroup)
+        await self.server.transport.leaveGroup(self.multicastGroup)
 
     # FIXME: https://twistedmatrix.com/trac/ticket/7780
     @skipIf(
@@ -688,7 +747,7 @@ class MulticastTests(TestCase):
         """
         # 127.0.0.1 is not a multicast address, so joining it should fail.
         return self.assertFailure(
-            self.client.transport.joinGroup("127.0.0.1"), error.MulticastJoinError
+            self.client.transport.joinGroup(self.invalidGroup), error.MulticastJoinError
         )
 
     def _test_multicast(self):
@@ -697,14 +756,14 @@ class MulticastTests(TestCase):
         received from it.
         """
         c = Server()
-        p = reactor.listenMulticast(0, c)
+        p = reactor.listenMulticast(0, c, self.interface)
         addr = self.server.transport.getHost()
 
-        joined = self.server.transport.joinGroup("225.0.0.250")
+        joined = self.server.transport.joinGroup(self.multicastGroup)
 
         def cbJoined(ignored):
             d = self.server.packetReceived = Deferred()
-            c.transport.write(b"hello world", ("225.0.0.250", addr.port))
+            c.transport.write(b"hello world", (self.multicastGroup, addr.port))
             return d
 
         joined.addCallback(cbJoined)
@@ -729,51 +788,80 @@ class MulticastTests(TestCase):
         "processes can listen, but not multiple sockets "
         "in same process?",
     )
-    def _test_multiListen(self):
+    async def _test_multiListen(self) -> None:
         """
         Test that multiple sockets can listen on the same multicast port and
         that they both receive multicast messages directed to that address.
         """
         firstClient = Server()
-        firstPort = reactor.listenMulticast(0, firstClient, listenMultiple=True)
-
-        portno = firstPort.getHost().port
+        mreactor = IReactorMulticast(reactor)
+        firstPort = mreactor.listenMulticast(
+            0, firstClient, listenMultiple=True, interface=self.interface
+        )
+        fpAddr = firstPort.getHost()
+        assert isinstance(fpAddr, (IPv4Address, IPv6Address))
+        portno = fpAddr.port
 
         secondClient = Server()
-        secondPort = reactor.listenMulticast(portno, secondClient, listenMultiple=True)
+        secondPort = mreactor.listenMulticast(
+            portno,
+            secondClient,
+            listenMultiple=True,
+            interface=self.interface,
+        )
 
-        theGroup = "225.0.0.250"
-        joined = gatherResults(
+        await gatherResults(
             [
-                self.server.transport.joinGroup(theGroup),
-                firstPort.joinGroup(theGroup),
-                secondPort.joinGroup(theGroup),
+                self.server.transport.joinGroup(self.multicastGroup),
+                firstPort.joinGroup(self.multicastGroup),
+                secondPort.joinGroup(self.multicastGroup),
             ]
         )
 
-        def serverJoined(ignored):
-            d1 = firstClient.packetReceived = Deferred()
-            d2 = secondClient.packetReceived = Deferred()
-            firstClient.transport.write(b"hello world", (theGroup, portno))
-            return gatherResults([d1, d2])
+        d1: Deferred[None]
+        d1 = firstClient.packetReceived = Deferred()
+        d2: Deferred[None]
+        d2 = secondClient.packetReceived = Deferred()
 
-        joined.addCallback(serverJoined)
+        firstClient.transport.write(b"hello world", (self.multicastGroup, portno))
+        await gatherResults([d1, d2])
 
-        def gotPackets(ignored):
-            self.assertEqual(firstClient.packets[0][0], b"hello world")
-            self.assertEqual(secondClient.packets[0][0], b"hello world")
+        self.assertEqual(firstClient.packets[0][0], b"hello world")
+        self.assertEqual(secondClient.packets[0][0], b"hello world")
 
-        joined.addCallback(gotPackets)
+        await gatherResults(
+            [
+                maybeDeferred(firstPort.stopListening),
+                maybeDeferred(secondPort.stopListening),
+            ]
+        )
 
-        def cleanup(passthrough):
-            result = gatherResults(
-                [
-                    maybeDeferred(firstPort.stopListening),
-                    maybeDeferred(secondPort.stopListening),
-                ]
-            )
-            result.addCallback(lambda ign: passthrough)
-            return result
 
-        joined.addBoth(cleanup)
-        return joined
+@skipIf(
+    not interfaces.IReactorMulticast(reactor, None),
+    "This reactor does not support multicast",
+)
+@skipWithoutIPv6
+@skipIf(
+    not checkMulticastAvailability(
+        AF_INET6, IPPROTO_IPV6, IPV6_JOIN_GROUP, "ff03::1", "::"
+    ),
+    "The local stack does not enable IPv6 multicast.",
+)
+class MulticastTestsIPv6(MulticastTests):
+    """
+    Tests for sending and receiving IPv6 multicast datagrams.
+    """
+
+    # Specifiy relevant IPv6 literals and values to override the IPv4 values in
+    # the superclass; all of these attributes are defined above.
+    interface: str = "::"
+    clientAddress: str = "::1"
+    multicastGroup: str = "ff03::1"
+    alternateInterface: str | int = next(
+        (idxnm[0] for idxnm in if_nameindex() if idxnm[1].startswith("lo"))
+    )
+    interfaceSynonym: str | int = alternateInterface
+    invalidGroup: str = "::1"
+    expectedInterface: str | int = 0
+    wrongAddressFamily: str = "127.0.0.1"
